-- t = time on clock
-- b = initial value
-- c = total change; target value - initial value
-- d = duration of the tween

-- a linear tween returns the total change multiplied by the clock, divided by duration of the tween, plus the initial value
local Easing = {}
function Easing.linearTween(t,b,c,d)
	return c*t/d + b;
end

function Easing.cubicOutTween(t,b,c,d)
	t = t/d;
	t = t - 1;
	return c*(t*t*t + 1) + b;
end

local tween_mt = {}
--[[
tween constructors
	tween(time,subject,target,easing,callback,arg1,arg2,...)

		Animates values in subject until equal to the key,value pairs specified in target, over time seconds.
		Optional easing, callback and arguments to the callback can also be specified.

		callback is a function that will be called when the tween has finished animating. Any additional parameters are forwarded to the callback function.

		tween returns a table with the following keys: duration, clock, subject, target, initial, easing, callback and arg, along with the tween_mt metatable. 
		(See tween methods below.)

		time 		float, duration of the tween in seconds
		subject		table, the object to animate
		target		table, a table of key-value pairs that will be animated in subject, i.e. {x=10} will animate x to the value 10 in subject.
		easing		function, an optional easing type for the tween. linearTween is the default.
					quad, cubic, quart, quint, expo, sine 	Different variations of smooth acceleration and deceleration
					back 	Causes the subject to overshoot and then pull back
					bounce 	Causes the subject to hit the target and bounce several times
					elastic	Causes the subject to oscilate like a guitar string around the target
					Each easing curve has four variants with the exception of linear
	
						in
						out
						inOut
						outIn
	
						Easing functions a combination of type and variant, such as 'elasticIn', 'quadInOut' and 'backOut'
	
						These easing functions were adapted from Emmanuel Oga's easing library
		callback	function, a function to call when the tween has completed

		Example

			box = { x = 10, y = 10, width = 20, height = 20, fillColor = {1,1,1,1} }
			tweens = {}
			tweens.a = tween(1, box, {fillColor = {1,0,0,1}})
			tweens.b = tween(1, box, {x = 100, y = 100}, quadIn)
			tweens.c = tween(1, box, {x = 100, y = 100}, linear)
			tweens.d = tween(1, box, {x = 100, y = 100}, linear, 
							 	  function()
									  print "Tween Ended"
							 	  end)
			tweens.e = tween(1, box, {x = 100, y = 100}, linear,
								  function(param)
							 		  print(param)
								  end, "Tween Ended")

			function love.update(dt)
				for k, tw in next, tweens do
					tw:update(dt)
				end
			end


	tween_delay(time,callback,arg1,arg2,...)
	tween_sequence(t1,t2,...)

note that tween.delay is currently a modified tween constructor using the same metatable as tween, so it's not as efficient as it should
tween_sequence takes multiple tweens and or sequences of tweens, and plays them back one after the other

tween methods (supported)
	tween.finish(tween)
	tween.update(tween, deltatime)
	tween.reset(tween)
	tween.resume(tween)
	tween.stop(tween)

sequence methods (supported)
	sequence.resetAll(sequence)
	sequence.resume(sequence)
	sequence.stop(sequence)
	sequence.update(sequence, deltatime)

The following methods look for tween metatables using assert or error. 
if you want to add your own tween (for example, a tween that ping-pongs forever), please edit each of the following functions
	tween_sequence (...)
	assert_tween_metatable (tween)
	sequence_mt.resetAll (sequence)

--]]

function tween_mt.finish(self) 
	self.clock=self.duration
	local t,b,c,d
	for k,v in pairs(self.target) do
		t,b,c,d = self.duration, self.initial[k], v - self.initial[k], self.duration
		self.subject[k] = self.easing(t,b,c,d)
	end
	if self.callback then self.callback(self.arg) end
end;

function tween_mt.update(self,dt)
	if self.stopped then return end
	if self.clock >= self.duration then
		self.clock=self.duration
		-- return true -- i put this here for sequences, but its not a big deal to check it there instead
	else
		local t,b,c,d
		self.clock=self.clock+dt
		if self.clock>=self.duration then
			for k,v in pairs(self.target) do
				t,b,c,d = self.duration, self.initial[k], v - self.initial[k], self.duration
				self.subject[k] = self.easing(t,b,c,d)
			end
			if self.callback then self.callback(self.arg) end
		else		
			for k,v in pairs(self.target) do
				t,b,c,d = self.clock, self.initial[k], v - self.initial[k], self.duration
				self.subject[k] = self.easing(t,b,c,d)
			end
		end
	end
end;

function tween_mt.stop(self)
	self.stopped = true
	-- Stops the tween. This does not reset the subject's values to their initial state.
end;

function tween_mt.reset(self)
	self.clock = 0
	for k,v in pairs(self.initial) do
		self.subject[k] = v
	end
	self.stopped = true
	-- Stops the tween and resets the subject's values to their initial state.
end;

function tween_mt.resetAll(self)
	error("called unsupported method 'resetAll' on tween instead of tween_sequence")
end;

function tween_mt.resume(self)
	self.stopped = false
	-- Resumes the tween. This does not reset the subject's values to their initial state.
end;


local _TWEENS = {}

local function updateTWEENS(dt) for k in next, _TWEENS do k:update(dt) end end

function tween_mt.register(self)
	if _TWEENS[self] then	error "Tween.register failed because _TWEEN[self] already exists"	end
	_TWEENS[self] = true
	-- Registers the tween. Uses the table 'self' as a key in _TWEEN. You can call methods using the key directly.
end

function tween_mt.deregister(self)
	if _TWEENS[self] then _TWEENS[self] = nil
	else error "Tween.deregister failed because _TWEENS[self] does not exist"
	end
end

tween_mt.__index = tween_mt




local sequence_mt = {}

function sequence_mt.update(self,dt)
	if self.stopped then return end
	if self.clock >= self.duration then
		self.clock = self.duration
	else
		self.tweens[self.selected]:update(dt)
		if self.tweens[self.selected].clock == self.tweens[self.selected].duration then
			self.selected = next(self.tweens, self.selected) or self.selected
		end
		self.clock = self.clock + dt
	end
end;

function sequence_mt.finish(self)
	error("called unsupported method 'finish' on tween_sequence instead of tween")
	-- There is no expected behaviour for finishing a sequence of tweens (or sequences of sequences of tweens)
	-- Possiblities for expected behaviour:
	-- 		Loop through self.tweens, setting each subject's values to their target state
	-- 		Call self.tweens[self.selected]:finish()
end;

function sequence_mt.finishAll(self)
	error("sequence.finishAll is not currently supported")
	-- Should this function finish all tweens:
	--# starting from the current selection?
	--# - or from the first tween in the array?
end;

function sequence_mt.stop(self)
	self.stopped = true
	-- Stops the tween_sequence. This does not reset the selected subject's values to their initial state.
end;

function sequence_mt.reset(self)
	error("called unsupported method 'reset' on tween_sequence instead of tween")
	-- There is no expected behaviour for resetting a sequence of tweens (or sequences of sequences of tweens)
	-- Possiblities for expected behaviour:
	-- 		Loop through self.tweens, resetting each subject's values to their target state
	-- 		Call self.tweens[self.selected]:reset()
	-- 		Set self.selected to the first index of the self.tweens array
end;

function sequence_mt.resetAll(self)
	for i = #self.selected, 1, -1 do
		local v = self.tweens[i]
		local mt = getmetatable(v)
		if mt==sequence_mt then
			v:resetAll()
		elseif mt==tween_mt then
			v:reset()
		else
			-- check if metatable is valid. this should only run if i forget to add a metatable or a user overwrites the metatable of a tween
			assert_tween_metatable(v)
		end
	end
	-- Resets each tween in tween_sequence. Starts at the last position and ends on the first
end;

function sequence_mt.resume(self)
	self.stopped = false
	-- Resumes the tween_sequence. This does not reset the subject's values to their initial state.

	if self.tweens[self.selected].stopped then
		-- sequence would hang if the selected tween is stopped, so i put in this error
		error("could not resume sequence: current selected is stopped")
	end
end;

sequence_mt.__index = sequence_mt

local function assert_tween_metatable(tween) local mt=getmetatable(tween) assert(mt==tween_mt or mt==sequence_mt, "got unexpected metatable: expected tween_mt, sequence_mt") end


local Tween = { easing = Easing }

Tween.updateRegisteredTweens = update_TWEENS

function Tween.sequence(...)
	local tweens = {...} 
	assert(tweens[1])
	local time = 0
	for i,v in ipairs(tweens) do
		assert_tween_metatable(v)
		time = time + v.duration
	end
	return setmetatable({
		duration=time,
		clock=0,
		tweens=tweens, 
		selected=1,
		stopped=false,
	}, sequence_mt)
end

function Tween.delay(time,callback,...)
	return setmetatable({
		duration=time,
		clock=0,
		callback=callback,
		subject={},
		target={},
		initial={},
		easing=Easing['linearTween'],
		arg=...,
		stopped=false,
	}, tween_mt)
end

--note that tween currently does not have looping defined
function Tween.new(time,subject,target,easing,callback,...)
	local initial={}
	for k in pairs(target) do initial[k] = subject[k] end
	return setmetatable({
		duration=time,
		clock=0,
		subject=subject,
		target=target,
		initial=initial,
		easing=Easing[easing] or Easing['linearTween'],
		callback=callback,
		arg=...,
		stopped=false,
	}, tween_mt)
end

return Tween
